#!/usr/bin/env python3
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
import teselagen


def load_from_json(filepath: Path) -> Any:
    """

    Loads a JSON file.

    Args:
        filepath (Path) : Path to the input JSON.

    Returns:
        (Any) : It returns a JSON object.

    """
    absolute_path: Path = filepath.absolute()

    json_obj: Any = json.loads(absolute_path.read_text())

    return json_obj

def get_project_root() -> Path:
    """ Returns project's root folder <absolute/path/to>/lib
    """
    return Path( teselagen.__path__[0] ).parent.resolve()

def get_credentials_path() -> Path:
    """ Returns path to where credentials file should be
    """
    return get_project_root() / '.credentials'

def get_test_configuration_path() -> Path:
    """ Returns path to where .test_configuration file should be
    """
    return get_project_root() / '.test_configuration'


#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

import json
import getpass
import pandas as pd
from os.path import join
from urllib.parse import urlencode, urlparse, urljoin
from typing import Any, Dict, List, Optional, TypeVar, Union

import requests

from teselagen.api.client import (DEFAULT_API_TOKEN_NAME, DEFAULT_HOST_URL,
                                  TeselaGenClient, get, post, put, requires_login,
                                  download_file)

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument


class DESIGNClient(TeselaGenClient):
    ALLOWED_SEQ_FORMATS = {'json', 'fasta', 'genbank'}
    URL_GET_ASSEMBLY_REPORT = "/assembly-report/export"

    def __init__(self,
                 api_token_name: str = DEFAULT_API_TOKEN_NAME,
                 host_url: str = DEFAULT_HOST_URL):
        module_name: str = "design"
        super(DESIGNClient, self).__init__(module_name=module_name,
                                           host_url=host_url,
                                           api_token_name=api_token_name)
        # EXPORT
        # GET
        # /export/sequence/{format}/{sequenceId}
        self.export_dna_sequence_url: str = f"{self.api_url_base}/export/sequence/"
        # GET
        # /export/sequences/{format}/
        self.export_dna_sequences_url: str = f"{self.api_url_base}/export/sequences/json"
        # GET
        # /export/aminoacids/{format}/{sequenceId}
        #self.export_aminoacid_sequence_url: str = f"{self.api_url_base}/export/aminoacids"

        ## IMPORT
        # POST
        #self.import_dna_sequence_url: str = f"{self.api_url_base}/import/sequence"
        # POST
        #self.import_aminoacid_sequence_url: str = f"{self.api_url_base}/import/aminoacids"
        ## DESIGN
        # GET
        # /designs/{id}
        self.get_design_url: str = f"{self.api_url_base}/designs"
        # DEL
        # /designs/{id}
        #self.delete_design_url: str = f"{self.api_url_base}/designs"
        # GET
        # /designs
        self.get_designs_url: str = f"{self.api_url_base}/designs"
        # POST
        # /designs
        #self.post_designs_url: str = f"{self.api_url_base}/designs"
        # POST
        # /codon-optimization-jobs
        self.post_codon_op: str = f"{self.api_url_base}/codon-optimization-jobs"
        # GET
        # /codon-optimization-jobs
        self.get_codon_op_result: str = f"{self.api_url_base}/codon-optimization-jobs"



        ## RBS Calculator API Tesealgen Integration endpoints

        # GET
        # /rbs-calculator/status
        self.rbs_calculator_status_url: str = f"{self.api_url_base}/rbs-calculator/status"

        # POST
        # /rbs-calculator/submit
        self.rbs_calculator_submit_url: str = f"{self.api_url_base}/rbs-calculator/submit"

        # GET
        # /rbs-calculator/jobs/
        self.rbs_calculator_jobs_url: str = join(self.api_url_base, "rbs-calculator/jobs")

        # GET
        # /rbs-calculator/jobs/:jobId
        self.rbs_calculator_job_url: str = join(self.api_url_base, "rbs-calculator/jobs") + "/{}"

        # GET
        # /rbs-calculator/organisms
        self.rbs_calculator_organisms_url: str = f"{self.api_url_base}/rbs-calculator/organisms"


    @requires_login
    def get_dna_sequence(self, seq_id: int, out_format:str='json', out_filepath:Optional[str]=None)->Union[str, dict]:
        """Gets full sequence record from its ID

        Args:
            seq_id (int): Sequence id (can be found within the sequence detail url at UI)
            out_format (str, optional): Output format. Use 'json' and the method
                will return a dict and 'fasta', or 'genbank' to return a string based
                on those formats. This also determines the output format when writing an
                output file. Defaults to 'json'.
            out_filepath (Optional[str], optional): Path to output file. If None it will
                not create any file. Defaults to None.

        Raises:
            ValueError: If format not available

        Returns:
            Union[str, dict]: A dict object with json data or a string if another format
                is chosen.
        """
        if out_format not in self.ALLOWED_SEQ_FORMATS:
            raise ValueError(f"Format {out_format} not in {self.ALLOWED_SEQ_FORMATS}")
        url = urljoin(self.export_dna_sequence_url, f'{out_format}/{seq_id}')
        response = get(url=url, headers=self.headers)
        # Write output file
        if out_filepath is not None:
            with open(out_filepath, 'w') as f:
                f.write(response["content"])
        # Parse json
        if out_format == 'json':
            response['content'] = json.loads(response["content"])
        # Finish
        return response["content"]

    @requires_login
    def get_dna_sequences(self, name: str)->List[dict]:
        """ Get all sequences which names matches a string

        Args:
            name (str): name of sequences to download

        Returns:
            List[dict]: [description]
        """
        args = {'name': name}
        response = get(url=self.export_dna_sequences_url,
                       headers=self.headers,
                       params=args)
        out = json.loads(response["content"])
        return out

    @requires_login
    def get_designs(self, name: Optional[str]=None, gql_filter: Optional[dict]=None)->List[dict]:
        """Retrieves a list of designs summary

        Args:
            name (str, optional):  Design's name
                to filter query. Defaults to None.
            gql_filter (dict, optional): GraphicQL
                filter dictionary. May be used to add additional filter conditions.
                See api-docs.Defaults to None.

        Returns:
            List[dict]: A list of designs info. Each dict in the list contains name and
                id of each design.
        """
        # Prepare parameters
        args: dict = {'gqlFilter':{}}
        if name is not None:
            args['gqlFilter']["name"] = name
        if gql_filter is not None:
            args['gqlFilter'].update(gql_filter)
        # Param gqlFilter should be a json string
        args['gqlFilter'] = json.dumps(args['gqlFilter'])
        # Make request and process output
        response = get(url=self.get_designs_url, headers=self.headers, params=args)
        out = json.loads(response["content"])
        # Remove unuseful key
        for el in out: el.pop("__typename")
        return out

    @requires_login
    def get_design(self, design_id: Union[int, str])->dict:
        """ Retrieves the design with specified id

        Raises error if design_id is not found

        Args:
            design_id (str, int):  Design's id

        Returns:
            dict: A dict containing designs information
        """
        response = get(url=f"{self.get_design_url}/{design_id}",
                       headers=self.headers)
                       #params=args)
        out = json.loads(response["content"])
        return out

    @requires_login
    def get_assembly_report(self, report_id: int, local_filename=None)->str:
        """
        Retrieves an assembly report given an id

        Args:
            report_id (int): The id of report as can be seen
                on the browser URL for that report view in the
                DESIGN module
        Returns:
            str path to output file

        """
        if local_filename is None:
            local_filename = f"report_{report_id}.zip"
        url = f"{self.api_url_base}{self.URL_GET_ASSEMBLY_REPORT}/{report_id}"
        return download_file(url=url, local_filename=local_filename, headers=self.headers)

    def post_codon_optimization_job(self, algorithm="ALGORITHMS_NAME", parameters={}):
        body = {
            "algorithm": algorithm,
            "parameters": parameters
        }
        response = post(url=self.post_codon_op,
                        headers=self.headers,
                        json=body)
        return json.loads(response["content"])

    def get_codon_optimization_job_results(self, job_id):
        response = get(url=f"{self.get_codon_op_result}/{job_id}",
                       headers=self.headers)
                       #params=args)
        out = json.loads(response["content"])
        return out


    ## RBS Calculator Methods.

    @requires_login
    def rbs_calculator_set_token(self, rbs_token: str = None)->None:
        """
        Sets TeselaGen-RBS calculator integration token.

        Args:
            rbs_token(str): Integration token. This is required to consume Tesealgen/RBS Calculator API. Please ask TeselaGen for your integration token.

        Returns:
            dict: {authenticated: boolean, success: boolean}
        """
        rbs_token = getpass.getpass(prompt="Enter x-tg-rbs-token: ") if rbs_token is None else rbs_token
        self.headers = {**self.headers, "x-tg-rbs-token": rbs_token}

    @requires_login
    def rbs_calculator_status(self)->dict:
        """
        Checks the status of the RBS Calculator Integration API.

        Returns:
            dict: {authenticated: boolean, success: boolean}
        """
        try:
            result = get(url=self.rbs_calculator_status_url, headers=self.headers)
        except Exception as e:
            return e

        return result["content"]

    @requires_login
    def rbs_calculator_get_jobs(self, job_id: str = None)->dict:
        """
        Fetches an RBS Calculator Job with the provided job_id.

        Args:
            job_id (str): ID of an RBS Calculator Job
        Returns:
            dict: {authenticated: boolean, success: boolean}
        """

        try:
            result = get(
                url=self.rbs_calculator_job_url.format(job_id) if job_id is not None else self.rbs_calculator_jobs_url,
                headers=self.headers)
        except Exception as e:
            return e

        return result["content"]

    @requires_login
    def rbs_calculator_organisms(self, as_dataframe: bool = True)->dict:
        """
        Fetches all available organisms or host supported by the RBS Calculator tools.

        Args:
            as_dataframe(bool): Whether to return the response as a dataframe.
        Returns:
            List[Dict[str, str]]: A list of all the available organisms/hosts with their names and NCBI Accession IDs.
        """

        try:
            result = get(url=self.rbs_calculator_organisms_url, headers=self.headers)
        except Exception as e:
            return e

        result = json.loads(result["content"])
        result = pd.DataFrame(result) if as_dataframe else result
        return result

    @requires_login
    def rbs_calculator_submit_job(self, algorithm: str, params: dict)->dict:
        """
        Submits a job to the RBS Calculator API Version v2.1. For deeper information on the RBS Calculator tools please refer to the following documentation:

        - Paper: https://www.researchgate.net/publication/51155303_The_Ribosome_Binding_Site_Calculator.
        - Browser Application: https://salislab.net/software/
        - Swagger API Documentation: https://app.swaggerhub.com/apis-docs/DeNovoDNA/JobControl/1.0.1


        The TeselaGen/RBS Integration currently supports one of the three following RBS Calculator Tools:

        - "ReverseRBS": Calls the RBS Calculator in Reverse Engineering mode to predict the translation
            initiation rate of each start codon in a mRNA sequence. ([Predict Translation Rates](https://salislab.net/software/predict_rbs_calculator))

            parameters:
                mRNA (str): Valid 'GATCU' mRNA sequence.
                long_UTR (boolean): Enables long UTRs.
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function)


        - "RBSLibraryCalculator_SearchMode": Calls the RBS Library Calculator in Search mode to design a ribosome binding site library
            to maximally cover a selected  translation rate space between a targeted minimum and maximum rate
            using the fewest number of RBS variants ([Optimize Expression Levels](https://salislab.net/software/design_rbs_library_calculator)).

            parameters:
                CDS (str): Valid 'GATCU' coding sequence.
                RBS_Constrains (str): Either an empty string or a valid degenerate nucleotide sequence ('GATCURYSWKMBDHVN').
                initial_RBS_sequence (str): Either an empty string or a valid 'GATCU' RBS sequence.
                    This is used to initialize ther RBS sequence exploration algorithm. If an empty string is provided,
                    a random RBS sequence will be used as the initilizing sequence.
                library_size (int): Number of RBS sequences in your library.
                maximum_consecutive_degeneracy (int): The maximum number of consecutive degeneracy nucleotides for the RBS library designs.
                minimum_translation_initiation_rate (int): Lowest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                maximum_translation_initiation_rate (int): Highest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function).
                pre_sequence (str): Either an empty string or a valid 'GATCU' mRNA sequence that is required to appear upstream (5') of the RBS sequence.


        - "RBSLibraryCalculator_GenomeSearchMode": Calls the RBS Library Calculator in Genome Editing mode to design a genomic ribosome binding site library
            to maximally cover a selected translation rate space between a targeted minimum and maximum rate,  while introducing the
            fewest number of consecutive genomic mutations. ([Optimize Expression Levels](https://salislab.net/software/design_rbs_library_calculator)).

            parameters:
                CDS (str): Valid 'GATCU' coding sequence.
                RBS_Constrains (str): Either an empty string or a valid degenerate nucleotide sequence ('GATCURYSWKMBDHVN').
                genomic_RBS_sequence (str): Genomic RBS sequence. Must be a valid 'GATCU' sequence.
                initial_RBS_sequence (str): Either an empty string or a valid 'GATCU' RBS sequence.
                    This is used to initialize ther RBS sequence exploration algorithm. If an empty string is provided,
                    a random RBS sequence will be used as the initilizing sequence.
                library_size (int): Number of RBS sequences in your library.
                maximum_consecutive_degeneracy (int): The maximum number of consecutive degeneracy nucleotides for the RBS library designs.
                minimum_translation_initiation_rate (int): Lowest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                maximum_translation_initiation_rate (int): Highest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function).
                pre_sequence (str): Either an empty string or a valid 'GATCU' mRNA sequence that is required to appear upstream (5') of the RBS sequence.


        Args:
            algorithm (str): This should be one for the three algorithm described above currently suppoprted by the TeselaGen/RBS Integration.
            params (dict): These are the parameters required by the chosen algorithms according to the RBS Calculator API Swagger specifications mentioned above.
                        For more information on the parameters meaning refer to the https://salislab.net/software/ browser application.

                        Examples for the tools parameter inputs are as folows:

                        'ReverseRBS' params:
                            {
                                "mRNA": "YOUR_mRNA_ SEQUENCE",
                                "long_UTR": false,
                                "organism": "Acetobacter pomorum"
                            }

                        'RBSLibraryCalculator_SearchMode' params:
                            {
                                "CDS": "YOUR_CDS_SEQUENCE",
                                "RBS_Constraints": 'TCTAGANNNNNNNNNNNNNNNNNNNNNNNNNGAATTC',
                                "initial_RBS_sequence": "GATTGCGTGTGAGTTCTGGCACGGAGGAGCACGTA",
                                "library_size": 16,
                                "maximum_consecutive_degeneracy": 6,
                                "maximum_translation_initiation_rate": 100,
                                "minimum_translation_initiation_rate": 10,
                                "organism": "Escherichia coli str. K-12 substr. MG1655",
                                "pre_sequence": ""
                            }
                        'RBSLibraryCalculator_GenomeSearchMode' params:
                            {
                                "CDS": "YOUR_CDS_SEQUENCE",
                                "RBS_Constraints": "",
                                "genomic_RBS_sequence": "CUCGUACGGUGCUAACGUGCUUAGU",
                                "initial_RBS_sequence": "",
                                "library_size": 16,
                                "maximum_consecutive_degeneracy": 6,
                                "maximum_translation_initiation_rate": 100,
                                "minimum_translation_initiation_rate": 10,
                                "organism": "Escherichia coli str. K-12 substr. MG1655",
                                "pre_sequence": ""
                            }

        Returns:
            JSON with RBS Calculator job response. This may depend on the chosen tool.
        """


        params = json.dumps({**params, **{"algorithm": algorithm}})
        try:
            result = post(url=self.rbs_calculator_submit_url, data=params, headers=self.headers)
        except Exception as e:
            return e

        result = json.loads(result["content"])
        return result



#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

import json
from typing import Any, Dict, List, Optional, TypeVar, Union, Tuple

import requests

from teselagen.api.client import (DEFAULT_API_TOKEN_NAME, DEFAULT_HOST_URL,
                                  TeselaGenClient, get, post, put, requires_login)

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument

ALLOWED_MODEL_TYPES: List[Union[str, None]] = [
    "predictive", "evolutive", "generative", None
]


class EVOLVEClient(TeselaGenClient):
    def __init__(self,
                 api_token_name: str = DEFAULT_API_TOKEN_NAME,
                 host_url: str = DEFAULT_HOST_URL):
        module_name: str = "evolve"
        super(EVOLVEClient, self).__init__(module_name=module_name,
                                           host_url=host_url,
                                           api_token_name=api_token_name)
        # Here we define the client endpoints
        # Example :
        #    self.some_endpoint_url: str = f"{self.api_url_base}/some_endpoint"
        self.create_model_url: str = f"{self.api_url_base}/create-model"

        self.get_model_url: str = f"{self.api_url_base}/get-model"
        self.get_models_by_type_url: str = f"{self.api_url_base}/get-models-by-type"
        self.get_model_datapoints_url: str = f"{self.api_url_base}/get-model-datapoints"

        self.submit_model_url: str = f"{self.api_url_base}/submit-model"
        self.delete_model_url: str = f"{self.api_url_base}/delete-model"
        self.cancel_model_url: str = f"{self.api_url_base}/cancel-model"

        self.get_models_url: str = f"{self.api_url_base}/get-models"
        self.get_completed_tasks_url: str = f"{self.api_url_base}/get-completed-tasks"

        self.crispr_guide_rnas_url: str = f"{self.api_url_base}/crispr-grnas"

    def _get_data_from_content(self, content_dict:dict)->dict:
        """Checks that an output dict from evolve endpoint is healthy, and returns the 'data' field

        Args:
            content_dict (dict): content field (as dictionary) from an api endpoint response

        Raises:
            IOError: If dictionary isn't healthy

        Returns:
            dict: data field from endpoint response
        """
        if content_dict["message"] != "Submission success.":
            raise IOError(f"A problem occured with query: {content_dict['message']}")
        if 'data' not in content_dict:
            raise IOError(f"Can`t found 'data' key in response: {content_dict}")
        return content_dict["data"]

    def get_model_info(self, model_id: int):
        """ Retrieves model general information

        This will return a JSON object with the metadata of a model filtered
        by the provided model ID.

        Args :
            model_id (int) :
                Model identifier.

        Returns :
            () : A dict containing model info. An example is shown below:

        ```
        {
            "id": "0",
            "labId": "1",
            "modelType": "predictive",
            "name": "My First Predictive Model",
            "description": "This is an example model",
            "status": "completed-successfully",
            "evolveModelInfo": {
                "microserviceQueueId":
                "1",
                "dataSchema": [{
                    "id": "1",
                    "name": "Descriptor1",
                    "value_type": "numeric",
                    "type": "descriptor"
                }, {
                    "id": "1",
                    "name": "Descriptor2",
                    "value_type": "numeric",
                    "type": "descriptor"
                }, {
                    "id": "2",
                    "name": "Target",
                    "value_type": "numeric",
                    "type": "target"
                }],
                "modelStats": {
                    "MAE": 45
                }
            }
        }
        ```

        """
        body = {"id": str(model_id)}
        response = post(url=self.get_model_url,
                        headers=self.headers,
                        json=body)
        response["content"] = json.loads(response["content"])
        # Check output
        return self._get_data_from_content(response["content"])

    def get_models_by_type(self, model_type: Optional[str] = None):
        """

        This will return a JSON object with the metadata of multiple models,
        filtered by the provided `model_type`.

        Args :
            model_type (str) :

        ```
            "predictive"
            "evolutive"
            "generative"
             None
        ```

        Returns :
            () :

        ```

        {
            "message":
            "Submission success.",
            "data": [{
                "id": "1",
                "labId": "1",
                "modelType": "evolutive",
                "name": "My First Evolutive Model",
                "description": "This is an example model",
                "status": "completed-successfully",
                "evolveModelInfo": {
                    "microserviceQueueId":
                    "1",
                    "dataSchema": [{
                        "id": "1",
                        "name": "Descriptor1",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "1",
                        "name": "Descriptor2",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "2",
                        "name": "Target",
                        "value_type": "numeric",
                        "type": "target"
                    }],
                    "modelStats": {
                        "MAE": 45
                    }
                }
            }, {
                "id": "2",
                "labId": "1",
                "modelType": "evolutive",
                "name": "My Second Evolutive Model",
                "description": "This is an example model",
                "status": "completed-successfully",
                "evolveModelInfo": {
                    "microserviceQueueId":
                    "1",
                    "dataSchema": [{
                        "id": "1",
                        "name": "Descriptor1",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "1",
                        "name": "Descriptor2",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "2",
                        "name": "Target",
                        "value_type": "numeric",
                        "type": "target"
                    }],
                    "modelStats": {
                        "MAE": 40
                    }
                }
            }]
        }

        ```

        """
        if model_type not in ALLOWED_MODEL_TYPES:
            raise ValueError(f"Type: {model_type} not in {ALLOWED_MODEL_TYPES}")
        # body = {"modelType": "null" if model_type is None else model_type}
        body = {"modelType": model_type}
        response = post(url=self.get_models_by_type_url,
                        headers=self.headers,
                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])


    def get_model_datapoints(self, model_id: int, datapoint_type: str,
                             batch_size: int, batch_number: int):
        """

        This will return a JSON object with an array of datapoints filtered by
        the provided model ID and datapoint type. This array will come in the
        data field in the response body. Each element of the array has a
        datapoint field, this corresponds to a JSON object with the datapoint
        data.

        Args :

            model_id (int) :
                ID of the model

            datapoint_type (str) :
                The `datapoint_type` has two options :

                    "input"
                    "output"

                One can fetch only input datapoints (a.k.a training datapoints)
                or just fetch the output datapoint (a.k.a predicted datapoints
                not seen in the training dataset).

            batch_size (int) :
                `batch_size` refers to the number of datapoints to fetch from
                the database table.

            batch_number (int) :
                `batch_number` depends on `batch_size`, and determines the
                index position offset of length `batch_size` from where to
                start fetching datapoints.

        Returns :

        ```
            {
                "message":
                "Submission success.",
                "data": [{
                    "modelId": "1",
                    "labId": "2",
                    "datapoint": {},
                    "datapointType": "input"
                }, {
                    "modelId": "1",
                    "labId": "2",
                    "datapoint": {},
                    "datapointType": "input"
                }, {
                    "modelId": "1",
                    "labId": "2",
                    "datapoint": {},
                    "datapointType": "input"
                }]
            }
        ```

        """

        body = {
            "modelId": str(model_id),
            "datapointType": datapoint_type,
            "batchSize": batch_size,
            "batchNumber": batch_number
        }

        response = post(url=self.get_model_datapoints_url,
                        headers=self.headers,
                        json=body)

        response["content"] = json.loads(response["content"])

        return response["content"]
        # raise NotImplementedError

    def submit_model(self,
                     data_input: List[Any],
                     data_schema: List[Any],
                     model_type: str,
                     configs: Optional[Any] = None,
                     name: str = "",
                     description: Optional[str] = None):
        """

        Submits a model for training.

        Args :
            data_input (List[Any]) :
                This is required and must contain a JSON array of JSON objects
                with the input training data.

                These objects must be consistent with the `data_schema`
                property.

        ```
                [{
                    "Descriptor1": "A0",
                    "Descriptor2": "B1",
                    "Target": "1"
                }, {
                    "Descriptor1": "A0",
                    "Descriptor2": "B2",
                    "Target": "2"
                }, {
                    "Descriptor1": "A0",
                    "Descriptor2": "B3",
                    "Target": "3"
                }]
        ```

            data_schema (List[Any]) :
                This is an array of the schema of the input data columns.

                The `name` property corresponds to the column's name.

                The `type` property determines whether the column is a "target"
                or a "descriptor" (feature). Only "target" and "descriptor"
                are supported.

                The `value_type` type determines the type of the column's
                values. Only "numeric" and "categoric" are supported.

        ```
                [{
                    "id": "1",
                    "name": "Descriptor1",
                    "value_type": "categoric",
                    "type": "descriptor"
                }, {
                    "id": "2",
                    "name": "Descriptor2",
                    "value_type": "categoric",
                    "type": "descriptor"
                }, {
                    "id": "3",
                    "name": "Target",
                    "value_type": "numeric",
                    "type": "target"
                }]
        ```
                - `id` : corresponds to the id (position) of the column in the
                    dataset.
                - `name` : corresponds to the name of the column (descriptor
                    or target)
                - `type` : describes whether the field is a descriptor
                    (feature) or a target.
                - `value_type` : defines the type of value of this column.
                    Available types are "numeric" or "categoric".

            model_type (str) :
                The type of model wanting to submit.

                Either "predictive", "evolutive" or "generative".

                NOTE : If submitting a "generative" model, there's no "target"
                       column, in fact there should only be one "descriptor"
                       column. This needs to be properly set in the dataSchema
                       field according to the documentation.

            configs (Optional[Any]) :
                This is an advanced property containing advanced configuration
                for the training execution. Please refer to Teselagen's Data
                Science Team.

            name (str) :
                This sets the Evolve Model's name.

            description (Optional[str]) :
                This gives the Evolve Model's a description.

        Returns :
            (dict) : A dictionary containing info of the submitted job. En example is shown below:

            ```
            {
                "authToken": "1d140371-a59f-4ad2-b57c-6fc8e0a20ff8",
                "checkInInterval": null,
                "controlToken": null,
                "id": "36",
                "input": {
                    "job": "modeling-tool",
                    "kwargs": {}
                },
                "lastCheckIn": null,
                "missedCheckInCount": null,
                "result": null,
                "resultStatus": null,
                "service": "ds-tools",
                "serviceUrl": null,
                "startedOn": null,
                "status": "created",
                "taskId": null,
                "trackingId": null,
                "completedOn": null,
                "createdAt": "2020-10-29T13:18:06.167Z",
                "updatedAt": "2020-10-29T13:18:06.271Z",
                "cid": null,
                "__typename": "microserviceQueue"
            }
            ```

        """
        body = {
            "dataInput": data_input,
            "dataSchema": data_schema,
            "modelType": model_type,
            "configs": {} if configs is None else configs,
            "name": name,
            "description": "" if description is None else description
        }
        response = post(url=self.submit_model_url,
                        headers=self.headers,
                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])


    def delete_model(self, model_id: int):
        """

        Deletes a model matching the specified `model_id`.

        Args :
            model_id (int) :
                The model id that wants to be deleted.

        Returns :
            () :

        """
        body = {"id": str(model_id)}
        response = post(url=self.delete_model_url,
                        headers=self.headers,
                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])
        # raise NotImplementedError

    def cancel_model(self, model_id: int):
        """

        Cancels the submission of a model matching the specified `model_id`.

        Args :
            model_id (int) :
                The model id that wants to be cancelled.

        Returns :
            () :

        """
        body = {"id": str(model_id)}
        response = post(url=self.cancel_model_url,
                        headers=self.headers,
                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])

    @requires_login
    def design_crispr_grnas(self,
                            sequence: str,
                            target_indexes: Optional[Tuple[int, int]]=None,
                            target_sequence: Optional[str]=None,
                            pam_site: str='NGG',
                            min_score: float=40.0,
                            max_number: Optional[int]=50):
        body = {
            'data':{
                'sequence': sequence},
            'options': {
                'pamSite': pam_site,
                'minScore': min_score}}
        if target_indexes is not None:
            body['data']['targetStart'] = target_indexes[0]
            body['data']['targetEnd'] = target_indexes[1]
        if target_sequence is not None:
            body['data']['targetSequence'] = target_sequence
        if max_number is not None:
            body['options']['maxNumber'] = max_number
        response = post(url=self.crispr_guide_rnas_url,
                        headers=self.headers,
                        json=body)
        return json.loads(response["content"])

    # def get_models(self):
    #     # POST
    #     """This will return an object of an evolveModel [[ IEvolveModelEntity ]] filtered by the provided model ID.c"""
    #     raise NotImplementedError

    # def create_model(self):
    #     # POST
    #     """

    #     This interface represents the standard request object for the
    #     CreateModel REST API. The following properties must be included in the
    #     REST request body.

    #     Args:
    #         taskInput: Any necessary inputs for the create model task (it can
    #             be an empty object).

    #         dataInput: This is required and must contain an array of objects
    #             with the input training data. These objects must be consistent
    #             with the dataSchema property.

    #         dataSchema: This is an array of the schema of the input data
    #             columns. The name property corresponds to the column's name.
    #             The type property says whether the column is a target or a
    #             descriptor (feature). The value_type property says the type of
    #             the column's values.

    #                         {
    #                             name	string
    #                             type	string
    #                             value_type	string
    #                         }

    #         configs: This is an advanced property containing advanced
    #             configuration for the training execution. Please refer to
    #             Teselagen's Data Science Python Library.

    #         name (str): This sets the Evolve Model's name.

    #         description (str): This gives the Evolve Model's a description.

    #     Returns:

    #     """

    #     body = {
    #         "taskInput": {},
    #         "dataInput": [{}],
    #         "dataSchema": [{
    #             "name": "string",
    #             "type": "string",
    #             "value_type": "string"
    #         }],
    #         "configs": {},
    #         "name":
    #         "string",
    #         "description":
    #         "string"
    #     }

    #     raise NotImplementedError

    # def get_completed_tasks(self):
    #     # GET
    #     raise NotImplementedError



#!/usr/bin/env python3
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pytest

from teselagen.api.test_client import TESTClient
from teselagen.utils import load_from_json

TEST_FILE_CONTENTS: str = r"""Line,Teselagen Example Descriptor 1,Teselagen Example Descriptor 2,Teselagen Example Target,Teselagen Example Target Metric
1,A0,B1,1,ug/mL
2,A0,B2,2,ug/mL
3,A0,B3,3,ug/mL
4,A0,B5,5,ug/mL
5,A0,B6,6,ug/mL
6,A0,B9,9,ug/mL
7,A0,B10,10,ug/mL
8,A0,B11,11,ug/mL
9,A0,B12,12,ug/mL
10,A0,B13,13,ug/mL"""

# TEST_FILE_CONTENTS: str = "Line,Teselagen Example Descriptor 1,Teselagen Example Descriptor 2,Teselagen Example Target,Teselagen Example Target Metric\n1,A0,B1,1,ug/mL\n2,A0,B2,2,ug/mL\n3,A0,B3,3,ug/mL\n4,A0,B5,5,ug/mL\n5,A0,B6,6,ug/mL\n6,A0,B9,9,ug/mL\n7,A0,B10,10,ug/mL\n8,A0,B11,11,ug/mL\n9,A0,B12,12,ug/mL\n10,A0,B13,13,ug/mL\n11,A0,B14,14,ug/mL\n12,A0,B15,15,ug/mL\n13,A0,B17,17,ug/mL\n14,A0,B18,18,ug/mL\n15,A1,B0,1,ug/mL\n16,A1,B2,3,ug/mL\n17,A1,B3,4,ug/mL\n18,A1,B5,6,ug/mL\n19,A1,B6,7,ug/mL\n20,A1,B7,8,ug/mL\n21,A1,B8,9,ug/mL\n22,A1,B11,12,ug/mL\n23,A1,B12,13,ug/mL\n24,A1,B13,14,ug/mL\n25,A1,B14,15,ug/mL\n26,A1,B15,16,ug/mL\n27,A1,B16,17,ug/mL\n28,A1,B17,18,ug/mL\n29,A1,B18,19,ug/mL\n30,A1,B19,20,ug/mL\n31,A2,B0,2,ug/mL\n32,A2,B1,3,ug/mL\n33,A2,B2,4,ug/mL\n34,A2,B3,5,ug/mL\n35,A2,B5,7,ug/mL\n36,A2,B6,8,ug/mL\n37,A2,B8,10,ug/mL\n38,A2,B9,11,ug/mL\n39,A2,B10,12,ug/mL\n40,A2,B12,14,ug/mL\n41,A2,B13,15,ug/mL\n42,A2,B14,16,ug/mL\n43,A2,B15,17,ug/mL\n44,A2,B16,18,ug/mL\n45,A2,B17,19,ug/mL\n46,A2,B18,20,ug/mL\n47,A3,B0,3,ug/mL\n48,A3,B2,5,ug/mL\n49,A3,B3,6,ug/mL\n50,A3,B4,7,ug/mL\n51,A3,B5,8,ug/mL\n52,A3,B6,9,ug/mL\n53,A3,B7,10,ug/mL\n54,A3,B9,12,ug/mL\n55,A3,B10,13,ug/mL\n56,A3,B11,14,ug/mL\n57,A3,B12,15,ug/mL\n58,A3,B13,16,ug/mL\n59,A3,B14,17,ug/mL\n60,A3,B15,18,ug/mL\n61,A3,B16,19,ug/mL\n62,A3,B17,20,ug/mL\n63,A3,B18,21,ug/mL\n64,A3,B19,22,ug/mL\n65,A4,B1,5,ug/mL\n66,A4,B2,6,ug/mL\n67,A4,B3,7,ug/mL\n68,A4,B4,8,ug/mL\n69,A4,B5,9,ug/mL\n70,A4,B6,10,ug/mL\n71,A4,B7,11,ug/mL\n72,A4,B8,12,ug/mL\n73,A4,B10,14,ug/mL\n74,A4,B11,15,ug/mL\n75,A4,B12,16,ug/mL\n76,A4,B13,17,ug/mL\n77,A4,B14,18,ug/mL\n78,A4,B16,20,ug/mL\n79,A4,B18,22,ug/mL\n80,A4,B19,23,ug/mL\n81,A5,B0,5,ug/mL\n82,A5,B1,6,ug/mL\n83,A5,B2,7,ug/mL\n84,A5,B3,8,ug/mL\n85,A5,B4,9,ug/mL\n86,A5,B6,11,ug/mL\n87,A5,B7,12,ug/mL\n88,A5,B9,14,ug/mL\n89,A5,B10,15,ug/mL\n90,A5,B11,16,ug/mL\n91,A5,B12,17,ug/mL\n92,A5,B13,18,ug/mL\n93,A5,B14,19,ug/mL\n94,A5,B15,20,ug/mL\n95,A5,B16,21,ug/mL\n96,A5,B17,22,ug/mL\n97,A5,B18,23,ug/mL\n98,A5,B19,24,ug/mL\n99,A6,B0,6,ug/mL\n100,A6,B1,7,ug/mL\n101,A6,B2,8,ug/mL\n102,A6,B6,12,ug/mL\n103,A6,B7,13,ug/mL\n104,A6,B8,14,ug/mL\n105,A6,B9,15,ug/mL\n106,A6,B10,16,ug/mL\n107,A6,B11,17,ug/mL\n108,A6,B12,18,ug/mL\n109,A6,B13,19,ug/mL\n110,A6,B14,20,ug/mL\n111,A6,B15,21,ug/mL\n112,A6,B17,23,ug/mL\n113,A6,B18,24,ug/mL\n114,A6,B19,25,ug/mL\n115,A7,B0,7,ug/mL\n116,A7,B2,9,ug/mL\n117,A7,B3,10,ug/mL\n118,A7,B4,11,ug/mL\n119,A7,B5,12,ug/mL\n120,A7,B6,13,ug/mL\n121,A7,B7,14,ug/mL\n122,A7,B9,16,ug/mL\n123,A7,B10,17,ug/mL\n124,A7,B11,18,ug/mL\n125,A7,B12,19,ug/mL\n126,A7,B14,21,ug/mL\n127,A7,B16,23,ug/mL\n128,A7,B18,25,ug/mL\n129,A7,B19,26,ug/mL\n130,A8,B0,8,ug/mL\n131,A8,B3,11,ug/mL\n132,A8,B5,13,ug/mL\n133,A8,B6,14,ug/mL\n134,A8,B7,15,ug/mL\n135,A8,B9,17,ug/mL\n136,A8,B10,18,ug/mL\n137,A8,B11,19,ug/mL\n138,A8,B12,20,ug/mL\n139,A8,B14,22,ug/mL\n140,A8,B15,23,ug/mL\n141,A8,B18,26,ug/mL\n142,A8,B19,27,ug/mL\n143,A9,B3,12,ug/mL\n144,A9,B4,13,ug/mL\n145,A9,B5,14,ug/mL\n146,A9,B6,15,ug/mL\n147,A9,B8,17,ug/mL\n148,A9,B10,19,ug/mL\n149,A9,B11,20,ug/mL\n150,A9,B12,21,ug/mL\n151,A9,B13,22,ug/mL\n152,A9,B14,23,ug/mL\n153,A9,B15,24,ug/mL\n154,A9,B16,25,ug/mL\n155,A9,B19,28,ug/mL\n156,A10,B0,10,ug/mL\n157,A10,B1,11,ug/mL\n158,A10,B3,13,ug/mL\n159,A10,B5,15,ug/mL\n160,A10,B7,17,ug/mL\n161,A10,B9,19,ug/mL\n162,A10,B10,20,ug/mL\n163,A10,B11,21,ug/mL\n164,A10,B12,22,ug/mL\n165,A10,B13,23,ug/mL\n166,A10,B14,24,ug/mL\n167,A10,B15,25,ug/mL\n168,A10,B17,27,ug/mL\n169,A10,B19,29,ug/mL\n170,A11,B0,11,ug/mL\n171,A11,B1,12,ug/mL\n172,A11,B2,13,ug/mL\n173,A11,B3,14,ug/mL\n174,A11,B4,15,ug/mL\n175,A11,B5,16,ug/mL\n176,A11,B6,17,ug/mL\n177,A11,B8,19,ug/mL\n178,A11,B10,21,ug/mL\n179,A11,B11,22,ug/mL\n180,A11,B12,23,ug/mL\n181,A11,B13,24,ug/mL\n182,A11,B15,26,ug/mL\n183,A11,B16,27,ug/mL\n184,A11,B17,28,ug/mL\n185,A11,B18,29,ug/mL\n186,A12,B0,12,ug/mL\n187,A12,B1,13,ug/mL\n188,A12,B2,14,ug/mL\n189,A12,B3,15,ug/mL\n190,A12,B5,17,ug/mL\n191,A12,B6,18,ug/mL\n192,A12,B7,19,ug/mL\n193,A12,B8,20,ug/mL\n194,A12,B9,21,ug/mL\n195,A12,B10,22,ug/mL\n196,A12,B11,23,ug/mL\n197,A12,B12,24,ug/mL\n198,A12,B14,26,ug/mL\n199,A12,B15,27,ug/mL\n200,A12,B17,29,ug/mL\n201,A12,B18,30,ug/mL\n202,A12,B19,31,ug/mL\n203,A13,B0,13,ug/mL\n204,A13,B1,14,ug/mL\n205,A13,B2,15,ug/mL\n206,A13,B3,16,ug/mL\n207,A13,B5,18,ug/mL\n208,A13,B6,19,ug/mL\n209,A13,B8,21,ug/mL\n210,A13,B9,22,ug/mL\n211,A13,B10,23,ug/mL\n212,A13,B11,24,ug/mL\n213,A13,B12,25,ug/mL\n214,A13,B15,28,ug/mL\n215,A13,B16,29,ug/mL\n216,A13,B18,31,ug/mL\n217,A13,B19,32,ug/mL\n218,A14,B1,15,ug/mL\n219,A14,B4,18,ug/mL\n220,A14,B5,19,ug/mL\n221,A14,B6,20,ug/mL\n222,A14,B9,23,ug/mL\n223,A14,B10,24,ug/mL\n224,A14,B11,25,ug/mL\n225,A14,B12,26,ug/mL\n226,A14,B13,27,ug/mL\n227,A14,B14,28,ug/mL\n228,A14,B15,29,ug/mL\n229,A14,B16,30,ug/mL\n230,A14,B17,31,ug/mL\n231,A14,B18,32,ug/mL\n232,A14,B19,33,ug/mL\n233,A15,B0,15,ug/mL\n234,A15,B1,16,ug/mL\n235,A15,B2,17,ug/mL\n236,A15,B3,18,ug/mL\n237,A15,B5,20,ug/mL\n238,A15,B6,21,ug/mL\n239,A15,B7,22,ug/mL\n240,A15,B9,24,ug/mL\n241,A15,B11,26,ug/mL\n242,A15,B12,27,ug/mL\n243,A15,B13,28,ug/mL\n244,A15,B14,29,ug/mL\n245,A15,B16,31,ug/mL\n246,A15,B17,32,ug/mL\n247,A15,B19,34,ug/mL\n248,A16,B0,16,ug/mL\n249,A16,B2,18,ug/mL\n250,A16,B3,19,ug/mL\n251,A16,B4,20,ug/mL\n252,A16,B5,21,ug/mL\n253,A16,B6,22,ug/mL\n254,A16,B7,23,ug/mL\n255,A16,B8,24,ug/mL\n256,A16,B9,25,ug/mL\n257,A16,B10,26,ug/mL\n258,A16,B11,27,ug/mL\n259,A16,B12,28,ug/mL\n260,A16,B13,29,ug/mL\n261,A16,B14,30,ug/mL\n262,A16,B15,31,ug/mL\n263,A16,B17,33,ug/mL\n264,A16,B18,34,ug/mL\n265,A16,B19,35,ug/mL\n266,A17,B0,17,ug/mL\n267,A17,B2,19,ug/mL\n268,A17,B4,21,ug/mL\n269,A17,B5,22,ug/mL\n270,A17,B6,23,ug/mL\n271,A17,B8,25,ug/mL\n272,A17,B9,26,ug/mL\n273,A17,B10,27,ug/mL\n274,A17,B11,28,ug/mL\n275,A17,B12,29,ug/mL\n276,A17,B13,30,ug/mL\n277,A17,B14,31,ug/mL\n278,A17,B15,32,ug/mL\n279,A17,B16,33,ug/mL\n280,A17,B17,34,ug/mL\n281,A17,B18,35,ug/mL\n282,A18,B0,18,ug/mL\n283,A18,B1,19,ug/mL\n284,A18,B2,20,ug/mL\n285,A18,B3,21,ug/mL\n286,A18,B4,22,ug/mL\n287,A18,B6,24,ug/mL\n288,A18,B7,25,ug/mL\n289,A18,B8,26,ug/mL\n290,A18,B9,27,ug/mL\n291,A18,B11,29,ug/mL\n292,A18,B12,30,ug/mL\n293,A18,B13,31,ug/mL\n294,A18,B14,32,ug/mL\n295,A18,B15,33,ug/mL\n296,A18,B16,34,ug/mL\n297,A18,B17,35,ug/mL\n298,A18,B18,36,ug/mL\n299,A19,B0,19,ug/mL\n300,A19,B1,20,ug/mL\n301,A19,B3,22,ug/mL\n302,A19,B4,23,ug/mL\n303,A19,B5,24,ug/mL\n304,A19,B6,25,ug/mL\n305,A19,B7,26,ug/mL\n306,A19,B9,28,ug/mL\n307,A19,B10,29,ug/mL\n308,A19,B11,30,ug/mL\n309,A19,B12,31,ug/mL\n310,A19,B13,32,ug/mL\n311,A19,B14,33,ug/mL\n312,A19,B16,35,ug/mL\n313,A19,B19,38,ug/mL"

# Default IDs for testing purposes :
#       experimentId = 1
#       parserId = 1


@pytest.mark.incremental
class TestTESTClient():
    @pytest.fixture
    def expiration_time(self) -> str:
        _expiration_time: str = "30m"
        return _expiration_time

    @pytest.fixture
    def headers(self) -> Dict[str, str]:
        _headers: Dict[str, str] = {"Content-type": "application/json"}
        return _headers

    @pytest.fixture
    def client(self, host_url: str, api_token_name: str) -> TESTClient:
        """

        A TEST client instance.

        Returns:
            (TESTClient) : An instance of the TEST client.

        """
        test_client = TESTClient(api_token_name=api_token_name,
                                 host_url=host_url)
        return test_client

    @pytest.fixture
    def logged_client(self, client: TESTClient,
                      expiration_time: str) -> TESTClient:
        """

        A logged TEST client instance.

        Returns:
            (TESTClient) : An instance of the TEST client.

        """
        client.login(#username=credentials["test_user"],
                     #passwd=credentials["test_password"],
                     expiration_time=expiration_time)
        return client

    @pytest.fixture
    def lab_id(self, client: TESTClient) -> int:
        """

        Get the lab id used for testing.

        Returns:
            (int) : The laboratory identifier used for testing.

        """
        available_labs = client.get_laboratories()
        _lab_id: int = available_labs[0]['id']
        # _lab_id: int = 1

        return _lab_id

    @pytest.fixture
    def select_laboratory(self, client: TESTClient, lab_id: int) -> TESTClient:
        """

        A selected lab to work on.
        """
        # available_labs = client.get_laboratories()
        # lab_id: int = available_labs[0]['id']
        client.select_laboratory(lab_id=lab_id)
        return

    @pytest.fixture
    def experiment_id(self) -> int:
        """

        Get the experiment id used for testing.

        Returns:
            (int) : The experiment identifier used for testing.

        """
        _experiment_id: int = 1
        return _experiment_id

    @pytest.fixture
    def assay_id(self) -> int:
        """

        Get the assay id used for testing.

        Returns:
            (int) : The assay identifier used for testing.

        """
        _assay_id: int = 1  # None
        return _assay_id

    @pytest.fixture
    def experiment(self, logged_client: TESTClient,
                   select_laboratory) -> Dict[str, Any]:
        client = logged_client
        experiment_name: str = "Python Test Client Experiment"
        experiment: Dict[str, Any] = client.create_experiment(
            experiment_name=experiment_name)
        return experiment

    @pytest.fixture
    def assay(self, logged_client: TESTClient,
              experiment: Dict[str, Any]) -> Dict[str, Any]:
        client = logged_client
        experiment_id: int = int(experiment["id"])
        assay_name: str = "Python Test Client Assay"
        # TODO : We may need to update this, probably with a parser or
        #        parser_id fixture
        parser_id: Optional[int] = None
        response: Dict[str, Any] = client.create_assay(
            experiment_id=experiment_id,
            assay_name=assay_name,
            parser_id=parser_id)
        return response

    def test_class_attributes(self) -> None:

        # We check if the class inherit the parents methods.

        parent_class_methods: List[str] = [
            "register", "login", "logout", "get_server_status", "create_token",
            "update_token", "get_api_info", "get_current_user",
            "get_laboratories", "select_laboratory", "unselect_laboratory"
        ]

        # We check if the class has the required methods.

        experiment_methods: List[str] = [
            "get_experiments", "create_experiment", "delete_experiment"
        ]

        assay_methods: List[str] = [
            "get_assays", "create_assay", "delete_assay"
        ]

        file_methods: List[str] = [
            "get_files_info", "upload_file", "download_file", "delete_file"
        ]

        metadata_methods: List[str] = [
            "get_metadata", "create_metadata", "delete_metadata"
        ]

        attributes: List[str] = [
            *parent_class_methods, *experiment_methods, *assay_methods,
            *file_methods, *metadata_methods
        ]

        assert all(hasattr(TESTClient, attribute) for attribute in attributes)

    def test_instance_attributes(self, client: TESTClient) -> None:

        # We check if the client inherit the required parents attributes.
        parent_class_attributes: List[str] = ["api_url_base", "labs_url"]

        # We check if the client has the required attributes.

        experiment_attributes: List[str] = [
            "get_experiments_url", "create_experiment_url",
            "delete_experiment_url"
        ]

        assay_attributes: List[str] = [
            "get_assays_url", "get_assays_by_experiment_url",
            "create_assay_url", "delete_assay_url"
        ]

        file_attributes: List[str] = [
            "get_files_info_url", "get_files_info_by_assay_url",
            "get_file_data_url", "delete_file_url", "upload_file_url",
            "upload_file_into_assay_url"
        ]

        metadata_attributes: List[str] = [
            "get_metadata_url", "create_metadata_url", "delete_metadata_url"
        ]

        attributes: List[str] = [
            *parent_class_attributes, *experiment_attributes,
            *assay_attributes, *file_attributes, *metadata_attributes
        ]

        assert all(hasattr(client, attribute) for attribute in attributes)

    def test_login(self, client: TESTClient, expiration_time: str, api_token_name) -> None:
        # Before login, the client has no tokens
        assert client.auth_token is None
        assert api_token_name not in client.headers.keys()

        # LOGIN
        client.login(#username=credentials["test_user"],
                     #passwd=credentials["test_password"],
                     expiration_time=expiration_time)

        # After login, the client has tokens
        assert isinstance(client.auth_token, str)
        assert api_token_name in client.headers.keys()
        assert isinstance(client.headers[api_token_name], str)

    def test_create_experiment(self, experiment: List[Dict[str, Any]]) -> None:
        response = experiment
        assert isinstance(response, dict)
        assert "id" in response.keys()
        assert response["id"] is not None

    def test_delete_experiment(self, logged_client: TESTClient,
                               experiment: List[Dict[str, Any]]) -> None:
        client = logged_client
        experiment_id = int(experiment["id"])
        response = client.delete_experiment(experiment_id=experiment_id)
        assert response is None

    def test_create_assay(self, assay: List[Dict[str, Any]]) -> None:
        response = assay
        assert isinstance(response, dict)
        assert "id" in response.keys()
        assert "name" in response.keys()
        assert response["id"] is not None

    def test_delete_assay(self, logged_client: TESTClient, lab_id: int,
                          assay: List[Dict[str, Any]]) -> None:
        client = logged_client
        assay_id: int = int(assay["id"])
        response = client.delete_assay(assay_id=assay_id)

        assert response is None

    def test_get_experiments(self, logged_client: TESTClient,
                             select_laboratory) -> None:
        client = logged_client
        response: List[Dict[str, Any]] = client.get_experiments()

        assert isinstance(response, list)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)
        assert all(key in element.keys() and element[key] is not None
                   for element in response for key in ["id", "name"])
        # assert all(element[key] is not None for element in response for key in ["id", "name"])

    @pytest.mark.skip(
        reason=
        "It returns an empty list. We probably need to create an assay first.")
    def test_get_assays(self, logged_client: TESTClient, experiment_id: int,
                        select_laboratory, assay) -> None:
        created_assay = assay
        client = logged_client

        response: List[Dict[str, Any]] = client.get_assays(
            experiment_id=experiment_id)

        assert isinstance(response, list)
        # NOTE: It may be an empty list (?)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)

        # TODO : We may have to test if an "experiment" object containing an
        #       "id" and "name" is included

        assert all(key in element.keys() and element[key] is not None
                   for element in response for key in ["id", "name"])
        # assert all( element[key] is not None for element in response for key in ["id", "name"])

    @pytest.mark.skip(
        reason="We should expect an 'assay' key that may be None.")
    def test_get_files(self, logged_client: TESTClient,
                       assay_id: Optional[int], select_laboratory) -> None:
        client = logged_client
        response: List[Dict[str, Any]] = client.get_files(assay_id=assay_id)

        assert isinstance(response, list)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)

        # TODO : We should expect an "assay" key that may be None
        assert all(key in element.keys() for element in response
                   for key in ["id", "name", "assay"])

        # TODO : We should expect an "assay" key that may be None
        assert all(element[key] is not None for element in response
                   for key in ["id", "name"])

    @pytest.mark.skip(reason="Test not finished")
    def test_upload_file(self, logged_client: TESTClient):
        client = logged_client

        filepath: str = ""
        assay_id: Optional[int] = None
        response = client.upload_file(filepath=filepath, assay_id=assay_id)

        assert response is not None

    @pytest.mark.skip(reason="Test not finished")
    def test_download_file(self, logged_client: TESTClient):
        client = logged_client

        file_id: int = 0
        response = client.download_file(file_id=file_id)

        assert response is not None

    @pytest.mark.skip(reason="Test not finished")
    def test_delete_file(self, logged_client: TESTClient):
        client = logged_client

        file_id: int = 0
        response = client.delete_file(file_id=file_id)

        assert response is None

    @pytest.mark.skip(reason="We need to create a test experiment first")
    def test_download_assay(self, logged_client, assay_id,
                            expected_downloaded_assay_length,
                            expected_downloaded_assay_row) -> None:

        client = logged_client

        # DOWNLOAD ASSAY
        full_list: Optional[bool] = None
        columns = None
        page_number: int = None
        page_size: int = None

        downloaded_assay = client.download_assay(assay_id=assay_id,
                                                 full_list=full_list,
                                                 columns=columns,
                                                 page_number=page_number,
                                                 page_size=page_size)

        # Since we are testing this, it may not be an empty list
        assert downloaded_assay != []

        assert len(downloaded_assay) == expected_downloaded_assay_length

        row = downloaded_assay[0]
        expected_row = expected_downloaded_assay_row

        assert isinstance(row, dict)
        assert all(isinstance(key, str) for key in row.keys())
        assert all(isinstance(value, str) for value in row.values())

        assert row == expected_row

    @pytest.mark.skip(
        reason="We don't know which experiment and parser IDs to use")
    def test_upload_assays(self, logged_client) -> None:

        client = logged_client

        # UPLOAD ASSAY
        filename: str = "test_data.csv"
        contents = TEST_FILE_CONTENTS
        assay_name: str = "test_name"
        experiment_id: str = 1
        parser_id: str = 1

        number_of_previous_assays = len(
            client.get_assays(experiment_id=experiment_id))

        uploaded_assay_id = client.upload_assay(filename=filename,
                                                contents=contents,
                                                experiment_id=experiment_id,
                                                parser_id=parser_id,
                                                assay_name=assay_name)

        assert isinstance(uploaded_assay_id, int)

        number_of_current_assays = len(
            client.get_assays(experiment_id=experiment_id))

        assert number_of_current_assays == number_of_previous_assays + 1

        # ASSAY LENGTH
        length_of_the_uploaded_assay: int = client.get_length_of_an_assay(
            assay_id=uploaded_assay_id)

        assert length_of_the_uploaded_assay == 9



__version__ = "20.52.0"

# TODO: Maybe figure out how to include the 'pyproject.toml' file as part of the package installation files.
# import toml

# PY_PROJECT_FILE = './pyproject.toml'

# parsed_tom = toml.load(PY_PROJECT_FILE)

# __version__ = parsed_tom['tool']['poetry']['version']

# print(__version__)




#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT
import getpass
import json
import time
from typing import Any, Dict, List, Optional, Tuple, Union
import requests
from pathlib import Path
from teselagen.utils import load_from_json, get_credentials_path

AVAILABLE_MODULES: List[str] = ["test", "evolve"]  # ["test", "learn"/"evolve"]
DEFAULT_HOST_URL: str = "https://platform.teselagen.com"
DEFAULT_API_TOKEN_NAME: str = "x-tg-cli-token"

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument


# TODO: Maybe is better to set a default value for expires_in = "30m" instead of "1d" (?) or 8 hours
class TeselaGenClient():
    """Python TeselaGen Client."""
    def __init__(self,
                 module_name: str,
                 host_url: str = DEFAULT_HOST_URL,
                 api_token_name: str = DEFAULT_API_TOKEN_NAME) -> None:
        """

        A Python Client to use for communication with the TeselaGen modules.

        Args :
            module (str) : The module name to use for communication.

                    Available Modules :
                                        "test"
                                        "evolve" (WIP),
                                        "design" (WIP),
                                        "build" (WIP)

            host_url (str) : The Host URL of the API.

                Default = "https://platform.teselagen.com"

            api_token_name (str) : The name of the API token to use.

                Default = "x-tg-cli-token"

        """
        # NOTE: Do not add passwords to the class attributes.
        #       Delete all passwords once they've been used.
        self.module_name: str = module_name
        self.host_url: str = host_url
        self.api_token_name: str = api_token_name

        # Here we define the Base URL.
        self.module_url: str = f"{self.host_url}/{self.module_name}"
        self.api_url_base: str = f"{self.module_url}/cli-api"

        # Here we define the client endpoints.
        self.register_url: str = f"{self.module_url}/register"
        self.login_url: str = f"{self.module_url}/login"
        self.info_url: str = f"{self.api_url_base}/info"
        self.status_url: str = f"{self.api_url_base}/public/status"
        self.auth_url: str = f"{self.api_url_base}/public/auth"

        # Laboratories
        # NOTE : Currently, laboratories information only exists in TEST,
        #        and its probably global.
        #  self.labs_url: str = f"{self.api_url_base}/laboratories"
        self.labs_url: str = f"{self.host_url}/test/cli-api/laboratories"

        # NOTE : The authorization token will be updated with the
        #        "login" method.
        self.auth_token: Optional[str] = None

        # Here we define the headers.
        self.headers: Dict[str, str] = {"Content-Type": "application/json"}

    def register(self, username: str, password: str):
        """
        Registers a new user.

        NB: Registering a new user might require ADMIN priviledges.
        """
        body={
          "email": username,
          "firstName": "test",
          "lastName": "user",
          "password": password,
          "passwordConfirm": password
        }
        response = post(url=self.register_url, json=body)
        response["content"] = json.loads(response["content"])
        return response

    def login(self,
              username: Optional[str] = None,
              password: Optional[str] = None,
              apiKey: Optional[str] = None,
              expiration_time: str = "1d") -> None:
        """

        Login to the CLI with the username used to login through the UI.
        A password or an apiKey is required. If none is provided password will be prompted.

        Args:
            username (Optional[str]) : A valid username (usually their email)
                to authenticate. If not provided, it will be prompted.

                Default : None

            password (Optional[str]) : A password for the user. If not provided
                it will be prompted.

                Default: None

            apiKey (Optional[str]) : An exclusive API password obtained from the TeselaGen Browser Application Settings.
                It has 1 day expiration.

                Default: None

            expiration_time (Optional[str]) : Expiration time for the
                authentication (token), in zeit/ms format.

                Default = "1d"

        """
        # NOTE: the apiKey is obtained as an alternative password with 1 day expiration.
        _password = apiKey if apiKey is not None else password
        username, password = get_credentials(username=username, password=_password)
        auth_token = self.create_token(username=username,
                                    password=password,
                                    expiration_time=expiration_time)
        del username, password
        # else:
        #     auth_token = apiKey
        # It will update the auth token and headers.
        self.update_token(token=auth_token)
        return None

    def logout(self,
               username: Optional[str] = None,
               password: Optional[str] = None) -> None:
        """
        Log out from the CLI. A password is required for comfirmation.

        Args:

            username (Optional[str]) : Username. If not provided, it will be
                prompted.

            password (Optional[str]) : Password. If not provided, it will be
                prompted.

        """
        # TODO : Implement a new endpoint to deauthorize a token.

        # We locally delete the last token.
        self.update_token(token=None)

        username, password = get_credentials(username=username, password=password)

        # We create a temporary token, and wait until it expires.
        _ = self.create_token(username=username,
                              password=password,
                              expiration_time="1s")

        del username, password

        # We wait (a few seconds) for the (temporary) token to expire
        time.sleep(3)

        # NOTE :Verify that the user is deauthorized after the return.
        # raise NotImplementedError
        return

    def get_server_status(self) -> str:
        """

        Gets the current Server Status.

        Returns:

        """
        response = get(url=self.status_url, headers=self.headers)

        return response["content"]

    def create_token(self, username: str, password: str,
                     expiration_time: str) -> Union[str, None]:
        """

        Create a new access token for the user with the given username,
        password and expiration time, and return the new access token.

        Args:

            username (str) : The username identifier to authenticate with the
                API.

            password (str) : The password identifier to authenticate with the
                API.


            expiration_time (str) : Expiration time for the authentication
                (token), in zeit/ms format.

                Example : "1d"

        Returns:
            (Union[str, None]) : It returns the authentication token (as a
                string) for the given email address, or None if the email
                address is not authenticated.

        """
        body = {
            "username": username,
            "password": password,
            "expiresIn": expiration_time
        }

        # This happens in the CLI
        try:
            response = put(url=self.auth_url, headers=self.headers, json=body)
        except Exception as e:
            # TODO : Use a logger
            print("Connection Refused")
            return None
        print("Connection Accepted")

        del username, password, body

        response["content"] = json.loads(response["content"])

        # TODO : We could log the expiration Date
        # expiration_date: str = content["expirationDate"]

        # NOTE: Should we raise an exception if the content is not a valid
        #       JSON ?
        token: Optional[
            str] = response["content"]["token"] if response["status"] else None

        return token

    # TODO : Rename this to update_class_token() or update_auth_token()
    def update_token(self, token: Optional[str]) -> None:
        """

        Update the authorization token in the class headers and class
        attributes.

        Args :
            token (Optional[str]) : The authorization token to update in
                headers and class attributes.

                If the token is None, it will locally delete the last token
                from the class attributes.

        """
        self.auth_token = token

        if self.auth_token is not None:
            # If a new token is provided, we update the headers
            self.headers[self.api_token_name] = self.auth_token
        else:
            # If the token provided is None, we remove the last token from the
            # headers.
            _ = self.headers.pop(
                self.api_token_name
            ) if self.api_token_name in self.headers.keys() else None

        return

    def get_api_info(self) -> str:
        """

        Gets the current info about CLI API.

        NOTE : To get the info, the client should be already authorized.

        Returns:

            (str) :

        """
        try:
            response = get(url=self.info_url, headers=self.headers)

        except Exception as e:
            # TODO: Verify if we need to raise an exception.
            response = {
                "url": self.info_url,
                "content": str(e),
                "status": False
            }

        return response["content"]

    def is_authorized(self):
        # TODO : Try with get_api_info()
        raise NotImplementedError

    def get_current_user(self):
        """

        Gets the current user based on the header token.

        Returns:
            ( ) :
        """
        # TODO : implement a method to get the expiration date of the current
        #        token
        response = get(url=self.auth_url, headers=self.headers)
        response["content"] = json.loads(response["content"])

        return response

    # Laboratories Endpoints

    def get_laboratories(self) -> List[Dict[str, Any]]:
        """

        Get all available laboratories for the current user.

        Returns :
            () : A list of laboratories objects.

        """
        response = get(url=self.labs_url, headers=self.headers)

        # response["content"] = [{"id" : str, "name": str}, ...]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def select_laboratory(self, lab_name: Optional[str]=None, lab_id: Optional[int]=None, ) -> None:
        """ Sets the selected laboratory and adds it to the instance headers.

        Changes the header from internal class state with the id of the selected lab.
        This method will raise an error if the identifier (lab_name or lab_id) is not
        found.

        Args:
            lab_name (str): The name of the lab. If not set, the method will
                use the lab_id parameter
            lab_id (int): ID of the lab. If not set the method will use the
                lab_name parameter as lab identifier
        """
        identifier = lab_name if lab_id is None else str(lab_id)
        search_field = 'name' if lab_id is None else 'id'
        if identifier is None:
            raise ValueError("Received None lab identifiers")
        labs = self.get_laboratories()
        lab = list(filter(lambda x: x[search_field]==identifier,labs))
        if len(lab)==0: raise IOError(
            f"Can't find {search_field} {identifier}. Available labs are {labs}")
        # Finally store labid in headers
        self.headers.update({"tg-active-lab-id": str(lab[0]['id'])})
        print(f"Selected Lab: {lab[0]['name']}")

    def unselect_laboratory(self) -> None:
        """ Clear the selection of a laboratory and removes it from instance headers."""
        del self.headers["tg-active-lab-id"]
        print(f"No Lab is now selected.")

    def get(self, url: str, **kwargs) -> Any:
        """

        Wrapper that already includes the instance headers on the request.

        """
        return get(url=url, headers=self.headers, **kwargs)

    def post(self, url: str, **kwargs) -> Any:
        """

        Wrapper that already includes the instance headers on the request.

        """
        return post(url=url, headers=self.headers, **kwargs)

    def delete(self, url: str, **kwargs) -> Any:
        """

        Wrapper that already includes the instance headers on the request.

        """
        return delete(url=url, headers=self.headers, **kwargs)

    def put(self, url: str, **kwargs) -> Any:
        """

        Wrapper that already includes the instance headers on the request.

        """
        return put(url=url, headers=self.headers, **kwargs)


def get_credentials(username: Optional[str] = None,
                    password: Optional[str] = None) -> Tuple[str, str]:
    """

    It prompts the user for credentials in case username/password aren't provided
    and credentials file wasn't found.

    Args:
        username (Optional[str]) :  A valid username address to authenticate.
            If not provided, it will be prompted.

            Default : None

        password (Optional[str]) : A password to authenticate with. If not
            provided it will be prompted.

            Default : None

    Returns:
        (Tuple[str, str]) : It returns the credentials as a tuple of strings,
            containing the username and password.

            (user, password)

    """
    # Check if crentials are defined on a file
    file_credentials = load_credentials_from_file()
    username = file_credentials[0] if username is None else username
    password = file_credentials[1] if password is None else password
    # If credentials aren't defined, get them from user input
    try:
        username = input(
            f"Enter username: ") if username is None else username
        password = getpass.getpass(
            prompt=f"Password for {username}: ") if password is None else password
    except IOError as e:
        msg = ("""There was an error with user input. If you are making parallel
               tests, make sure you are avoiding 'input' by adding CREDENTIALS
               file.""")
        raise IOError(msg)
    # End
    return username, password

def load_credentials_from_file(path_to_credentials_file: str=None)->Tuple[Optional[str], Optional[str]]:
    """Load credentials from json credentials file

    The credentials file should contain a JSON object
    with the following keys (and the values)

    ```
    {
        "username": "user",
        "password": "password"
    }
    ```

    Args:
        path_to_credentials_file (str): Path to the file. If not set it will check for `.credentials` file
            at the folder that holds this method.
    Returns:
        username, password: Username and password strings if info is found in a credentials file, and (None, None)
            if not.
    """
    if path_to_credentials_file is None:
        path_to_credentials_file = str(get_credentials_path())
    if not Path(path_to_credentials_file).is_file():
        return None, None
    credentials: Dict = load_from_json(filepath=Path(path_to_credentials_file))
    return credentials['username'], credentials['password']

def handler(func):
    """

    Decorator to handle the response from a request.

    """
    def wrapper(**kwargs):
        # -> requests.Response
        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs.pop("url")

        try:
            response: requests.Response = func(url, **kwargs)

            if response.ok:
                return response

            elif response.status_code == 400:
                resp = json.loads(response.content)
                message: str = f"{response.reason}: {resp['error']}"
                raise Exception(message)

            elif response.status_code == 401:
                message: str = f"URL : {url} access is unauthorized."
                raise Exception(message)

            elif response.status_code == 404:
                message: str = f"URL : {url} cannot be found."
                raise Exception(message)

            elif response.status_code == 405:
                message: str = f"Method not allowed. URL : {url}"
                raise Exception(message)

            # TODO : Add more exceptions.

            else:
                # reason: str = response.reason
                message: str = f"Got code : {response.status_code}. Reason : {response.reason}"
                raise Exception(message)

        except Exception as e:
            raise

    return wrapper


def parser(func):
    """

    Decorator to parse the response from a request.

    """
    def wrapper(**kwargs) -> Dict[str, Union[str, bool, None]]:

        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs["url"]

        response: requests.Response = func(**kwargs)

        # TODO : Should we get/return JSON Serializables values ?
        # status 204 has no content.
        if response.status_code == 204:
            print("Deletion successful.")
            return

        response_as_json: Dict[str, Union[str, bool, None]] = {
            "url": url,
            "status": response.ok,
            "content": response.content.decode() if response.ok else None
        }

        return response_as_json

    return wrapper

def requires_login(func):
    """ Decorator to perform login beforehand, if necessary

    Add this decorator to any function from Client or a children
    that requires to be logged in.
    """
    def wrapper(self, *args, **kwargs):
        if self.auth_token is None:
            self.login()
            if self.auth_token is None:
                raise Exception("Could not access API, access token missing. Please use the 'login' function to obtain access.")
        return func(self, *args, **kwargs)
    return wrapper


@parser
@handler
def get(url: str, params: dict=None, **kwargs):
    """

    Same arguments and behavior as requests.get but handles exceptions and
    returns a dictionary instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Returns:
        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.get(url, params=params, **kwargs)
    return response


@parser
@handler
def post(url: str, **kwargs) -> requests.Response:
    """

    Same as requests.post but handles exceptions and returns a dictionary
    instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Example :

        url = "https://www.some_url.com/"
        response = post(url=url)

    Wrong usage:

        url = "https://www.some_url.com/"
        response = post(url)

    Returns:

        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.post(url, **kwargs)
    return response


@parser
@handler
def delete(url: str, **kwargs) -> requests.Response:
    """
    Same as requests.delete but handles exceptions and returns a dictionary
    instead of a requests.Response.

    """
    response: requests.Response = requests.delete(url, **kwargs)
    return response


@parser
@handler
def put(url: str, **kwargs):
    response: requests.Response = requests.put(url, **kwargs)
    return response


def download_file(url: str, local_filename: str=None, **kwargs)->str:
    """ Downloads a file from the specified url
    """
    if local_filename is None:
        local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    chunk_size = None
    with requests.get(url, stream=True, **kwargs) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                if chunk:
                    f.write(chunk)
    return local_filename
